# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
File search methods.
"""

import abc
import glob
import re
import warnings
from pathlib import Path
from datetime import timedelta
from datetime import datetime
from collections import defaultdict

import numpy as np

from tqdm import tqdm
from tqdm.dask import TqdmCallback

from dask.delayed import delayed
from dask.base import compute

class FilenameTemplate:
    """
    FilenameTemplate class.
    """

    def __init__(self, root_path, fn_templ, sf_templ=None):
        """
        Initialize root path, filename pattern and optional subfolder pattern.

        Parameters
        ----------
        root_path : str
            Root path.
        fn_templ : str
            Filename (glob) pattern.
            e.g. "{date}_*.{suffix}"
        sf_templ : dict, optional
            Subfolder pattern defined as dictionary (default: None).
            Keys represent unique meta names of subfolders and values define
            real folder names and/or (glob) pattern.
            e.g. {"variable": "*", "tile": "EN012*"}
        """
        self.root_path = Path(root_path)
        self.fn_templ = fn_templ
        self.sf_templ = sf_templ

    @property
    def template(self):
        """
        Name property.
        """
        if self.sf_templ is None:
            filename = self.root_path / self.fn_templ
        else:
            filename = self.root_path.joinpath(*list(self.sf_templ.values()),
                                               self.fn_templ)

        return filename

    def build_filename(self, fn_fmt, sf_fmt=None):
        """
        Create filename from format dictionary.

        Parameters
        ----------
        fn_fmt : dict
            Filename format applied on filename pattern (fn_pattern).
            e.g. fn_pattern = "{date}*.{suffix}"
            with fn_format_dict = {"date": "20000101", "suffix": "nc"}
            returns "20000101*.nc"
        sf_fmt : dict of dicts
            Format dictionary for subfolders. Each subfolder contains
            a dictionary defining the format of the folder name.
            e.g. sf_templ = {"years": {year}, "months": {month}}
            with sf_format = {"years": {"year": "2000"},
                                "months": {"month": "02"}}
            returns ["2000", "02"]

        Returns
        -------
        filename : str
            Filename with format_dict applied.
        """
        fn = self.build_basename(fn_fmt)

        if sf_fmt is None:
            filename = self.root_path / fn
        else:
            sf = self.build_subfolder(sf_fmt)
            filename = self.root_path.joinpath(*sf, fn)

        return str(filename)

    def build_basename(self, fmt):
        """
        Create file basename from format dictionary.

        Parameters
        ----------
        fmt : dict
            Filename format applied on filename pattern (fn_pattern).
            e.g. fn_pattern = "{date}*.{suffix}"
            with fmt = {"date": "20000101", "suffix": "nc"}
            returns "20000101*.nc"

        Returns
        -------
        filename : str
            Filename with format_dict applied.
        """
        return self.fn_templ.format(**fmt)

    def build_subfolder(self, fmt):
        """
        Create subfolder path from format dictionary.

        Parameters
        ----------
        fmt : dict of dicts
            Format dictionary for subfolders. Each subfolder contains
            a dictionary defining the format of the folder name.
            e.g. sf_pattern = {"years": {year}, "months": {month}}
            with format_dict = {"years": {"year": "2000"},
                                "months": {"month": "02"}}
            returns ["2000", "02"]

        Returns
        -------
        subfolder : list of str
            Subfolder with format_dict applied.
        """
        subfolder = []

        if self.sf_templ is not None:
            for name, v in self.sf_templ.items():
                if fmt is not None:
                    if name in fmt:
                        subfolder.append(
                            self.sf_templ[name].format(**fmt[name]))
                    else:
                        subfolder.append(self.sf_templ[name])
                else:
                    subfolder.append(self.sf_templ[name])

        return subfolder


class FileSearch:
    """
    FileSearch class.
    """

    def __init__(self, root_path, fn_pattern, sf_pattern=None):
        """
        Initialize root path, filename pattern and optional subfolder pattern.

        Parameters
        ----------
        root_path : str
            Root path.
        fn_pattern : str
            Filename (glob) pattern.
            e.g. "{date}_*.{suffix}"
        sf_pattern : dict, optional
            Subfolder pattern defined as dictionary (default: None).
            Keys represent unique meta names of subfolders and values define
            real folder names and/or (glob) pattern.
            e.g. {"variable": "*", "tile": "EN012*"}
        """
        self.file_templ = FilenameTemplate(root_path, fn_pattern, sf_pattern)

    def search(self, fn_fmt, sf_fmt=None, recursive=False):
        """
        Search filesystem for given pattern returning list.

        Parameters
        ----------
        fn_fmt : dict
            Filename format dictionary.
        sf_fmt : dict of dicts, optional
            Format dictionary for subfolders (default: None).
        recursive : bool, optional
            If recursive is true, the pattern "**" will match any files and
            zero or more directories, subdirectories and symbolic links to
            directories (default: False).

        Returns
        -------
        filenames : list of str
            Return a possibly-empty list of path/file names that match.
        """
        return glob.glob(
            self.file_templ.build_filename(fn_fmt, sf_fmt),
            recursive=recursive)

    def isearch(self, fn_fmt, sf_fmt=None, recursive=False):
        """
        Search filesystem for given pattern returning iterator.

        Parameters
        ----------
        fn_fmt : dict
            Filename format dictionary.
        sf_fmt : dict of dicts, optional
            Format dictionary for subfolders (default: None).
        recursive : bool, optional
            If recursive is true, the pattern "**" will match any files and
            zero or more directories, subdirectories and symbolic links to
            directories (default: False).

        Returns
        -------
        filenames : iterator
            Iterator which yields the same values as search() without
            actually storing them all simultaneously.
        """
        return glob.iglob(
            self.file_templ.build_filename(fn_fmt, sf_fmt),
            recursive=recursive)

    def create_search_func(self, func, recursive=False):
        """
        Create custom search function returning it.

        Parameters
        ----------
        func : function
            Search function with its own args/kwargs returning
            a filename format dictionary and subfolder format dictionary
            depending on the passed arguments.
        recursive : bool, optional
            If recursive is true, the pattern "**" will match any files and
            zero or more directories, subdirectories and symbolic links to
            directories (default: False).

        Returns
        -------
        custom_search : function
            Custom search function returning a possibly-empty list
            of path/file names that match.
        """

        def custom_search(*args, **kwargs):
            fn_fmt, sf_fmt = func(*args, **kwargs)
            return self.search(fn_fmt, sf_fmt, recursive=recursive)

        return custom_search

    def create_isearch_func(self, func, recursive=False):
        """
        Create custom search function returning it.

        Parameters
        ----------
        func : function
            Search function with its own args/kwargs returning
            a filename format dictionary and subfolder format dictionary
            depending on the passed arguments.
        recursive : bool, optional
            If recursive is true, the pattern "**" will match any files and
            zero or more directories, subdirectories and symbolic links to
            directories (default: False).

        Returns
        -------
        custom_search : function
            Custom search function returning an iterator of path/file names
            that match.
        """

        def custom_search(*args, **kwargs):
            fn_fmt, sf_fmt = func(*args, **kwargs)
            return self.isearch(fn_fmt, sf_fmt, recursive=recursive)

        return custom_search


class MultiFileHandler(metaclass=abc.ABCMeta):
    """
    MultiFileHandler class.
    """

    def __init__(
        self,
        root_path,
        cls,
        fn_templ,
        sf_templ=None,
        cls_kwargs=None,
        err=False,
        cache_size=0,
    ):
        """
        Initialize MultiFileHandler class.

        Parameters
        ----------
        root_path : str
            Root path.
        cls : class
            Class reading/writing files.
        fn_templ : str
            Filename template (e.g. "{date}_ascat.nc").
        sf_templ : dict, optional
            Subfolder template defined as dictionary (default: None).
        cls_kwargs : dict, optional
            Class keyword arguments (default: None).
        err : bool, optional
            Set true if a file error should be re-raised instead of
            reporting a warning.
            Default: False
        cache_size : int, optional
            Number of files to keep in memory (default=0).
        """
        self.root_path = root_path
        self.cls = cls
        self.ft = FilenameTemplate(root_path, fn_templ, sf_templ)
        self.fid = None
        self.err = err

        self.cache_size = cache_size
        if cache_size > 0:
            self.cache = {}

        if cls_kwargs is None:
            self.cls_kwargs = {}
        else:
            self.cls_kwargs = cls_kwargs

        self.fs = FileSearch(self.root_path, self.ft.fn_templ,
                             self.ft.sf_templ)

    def __enter__(self):
        """
        Context manager initialization.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Exit the runtime context related to this object. The file will be
        closed. The parameters describe the exception that caused the
        context to be exited.
        """
        self._close()

    def _open(self, filename):
        """
        Open file, i.e. create io class instance.

        Parameters
        ----------
        filename : str
            Filename.
        """
        self._close()

        try:
            self.fid = self.cls(filename, **self.cls_kwargs)
        except IOError:
            self.fid = None
            if self.err:
                raise
            else:
                warnings.warn(f"IOError: {filename}")

    def _close(self):
        """
        Try closing file.
        """
        if self.fid is not None and hasattr(self.fid, "close"):
            self.fid.close()
            self.fid = None

    @abc.abstractmethod
    def _fmt(*args, **kwargs):
        """
        Filename format and subfolder format used to read/write
        individual files.

        Returns
        -------
        fn_read_fmt : dict
            Filename format.
        sf_read_fmt : dict
            Subfolder format.
        fn_write_fmt : dict
            Filename format.
        sf_write_fmt : dict
            Subfolder format.
        """
        return

    def read(self, *fmt_args, fmt_kwargs=None, cls_kwargs=None):
        """
        Read data.

        Parameters
        ----------
        fmt_args : tuple
            Format arguments.
        fmt_kwargs : dict, optional
            Format keywords (Default: None).
        cls_kwargs : dict, optional
            Class keywords (Default: None).

        Returns
        -------
        data : dict, numpy.ndarray
            Data stored in file.
        """
        if fmt_kwargs is None:
            fmt_kwargs = {}

        if cls_kwargs is None:
            cls_kwargs = {}

        fn_read_fmt, sf_read_fmt, _, _ = self._fmt(*fmt_args, **fmt_kwargs)
        search_filename = self.ft.build_filename(fn_read_fmt, sf_read_fmt)
        filename = glob.glob(search_filename)
        self._open(filename)

        data = None
        data = self.fid.read(**cls_kwargs)

        return data

    def write(self, data, *fmt_args, fmt_kwargs=None, cls_kwargs=None):
        """
        Write data.

        Parameters
        ----------
        data : dict, numpy.ndarray
            Data to write.
        fmt_args : tuple
            Format arguments.
        fmt_kwargs : dict, optional
            Format keywords (Default: None).
        cls_kwargs : dict, optional
            Class keywords (Default: None).
        """
        if fmt_kwargs is None:
            fmt_kwargs = {}

        if cls_kwargs is None:
            cls_kwargs = {}

        _, _, fn_write_fmt, sf_write_fmt = self._fmt(*fmt_args, **fmt_kwargs)
        filename = self.ft.build_filename(fn_write_fmt, sf_write_fmt)
        self.write_file(data, filename, cls_kwargs=cls_kwargs)

    def read_file(self, filename, cls_kwargs=None):
        """
        Read data for given filename.

        Parameters
        ----------
        filename : str
            Filename.
        """
        if self.cache_size > 0 and filename in self.cache:
            return self.cache[filename]

        self._open(filename)
        data = self.fid.read(**cls_kwargs)

        if self.cache_size > 0:
            if len(self.cache) == self.cache_size:
                del self.cache[next(iter(self.cache))]
            self.cache[filename] = data

        return data

    def write_file(self, data, filename, cls_kwargs=None):
        """
        Write data for given filename.

        Parameters
        ----------
        filename : str
            Filename.
        """
        Path(filename).parent.mkdir(parents=True, exist_ok=True)

        self._open(filename)
        self.fid.write(data, **cls_kwargs)

    def search(
        self,
        fn_search_pattern,
        sf_search_pattern=None,
        custom_fn_templ=None,
        custom_sf_templ=None,
    ):
        """
        Search files for given root path and filename/folder pattern.

        Returns
        -------
        filenames : list of str
            Filenames.
        """
        fn_templ = custom_fn_templ if custom_fn_templ else self.ft.fn_templ
        sf_templ = custom_sf_templ if custom_sf_templ else self.ft.sf_templ

        if custom_fn_templ or custom_sf_templ:
            self.fs = FileSearch(self.root_path, fn_templ, sf_templ)

        filenames = self.fs.search(fn_search_pattern, sf_search_pattern)

        return sorted(filenames, reverse=True)


class ChronFiles(MultiFileHandler):
    """
    Managing chronological files with a date field in the filename.
    """

    def __init__(
        self,
        root_path,
        cls,
        fn_templ,
        sf_templ,
        cls_kwargs=None,
        err=True,
        fn_read_fmt=None,
        sf_read_fmt=None,
        fn_write_fmt=None,
        sf_write_fmt=None,
        cache_size=0,
    ):
        """
        Initialize ChronFiles class.

        Parameters
        ----------
        root_path : str
            Root path.
        cls : class
            Class reading/writing files.
        fn_templ : str
            Filename template (e.g. "{date}_ascat.nc").
        sf_templ : dict, optional
            Subfolder template defined as dictionary (default: None).
        cls_kwargs : dict, optional
            Class keyword arguments (default: None).
        err : bool, optional
            Set true if a file error should be re-raised instead of
            reporting a warning.
            Default: False
        fn_read_fmt : str or function, optional
            Filename format for read operation.
        sf_read_fmt : str or function, optional
            Subfolder format for read operation.
        fn_write_fmt : str or function, optional
            Filename format for write operation.
        sf_write_fmt : str or function, optional
            Subfolder format for write operation.
        cache_size : int, optional
            Number of files to keep in memory (default=0).
        """
        super().__init__(root_path, cls, fn_templ, sf_templ, cls_kwargs, err,
                         cache_size)

        self.fn_read_fmt = fn_read_fmt
        self.sf_read_fmt = sf_read_fmt
        self.fn_write_fmt = fn_write_fmt
        self.sf_write_fmt = sf_write_fmt

    def _fmt(self, *fmt_args, **fmt_kwargs):
        """
        Format filenames/filepaths.
        """
        if callable(self.fn_read_fmt):
            fn_read_fmt = self.fn_read_fmt(*fmt_args, **fmt_kwargs)
        else:
            fn_read_fmt = self.fn_read_fmt

        if callable(self.sf_read_fmt):
            sf_read_fmt = self.sf_read_fmt(*fmt_args, **fmt_kwargs)
        else:
            sf_read_fmt = self.sf_read_fmt

        if callable(self.fn_write_fmt):
            fn_write_fmt = self.fn_write_fmt(*fmt_args, **fmt_kwargs)
        else:
            fn_write_fmt = self.fn_write_fmt

        if callable(self.sf_write_fmt):
            sf_write_fmt = self.sf_write_fmt(*fmt_args, **fmt_kwargs)
        else:
            sf_write_fmt = self.sf_write_fmt

        return fn_read_fmt, sf_read_fmt, fn_write_fmt, sf_write_fmt

    def _parse_date(self, filename, date_field, date_field_fmt):
        """
        Parse datetime from filename.

        Parameters
        ----------
        filename : str
            Filename.

        Returns
        -------
        timestamp : datetime
            File timestamp.
        """
        # escape special characters in the template string
        escaped_template = re.escape(self.ft.fn_templ)

        # replace escaped curly braces with capturing groups
        pattern = re.sub(r'\\{(.*?)\\}', r'(?P<\1>.*?)', escaped_template)

        match = re.match(pattern, Path(filename).name)
        date_substring = match.group(date_field)
        date = datetime.strptime(date_substring, date_field_fmt)

        return date

    def _merge_data(self, data, **kwargs):
        """
        Merge datasets after reading period. Needs to be overwritten
        by child class, otherwise data is returned as is.

        Parameters
        ----------
        data : list
            Data.
        **kwargs : dict
            Additional keyword arguments to the fid's merge method.

        Returns
        -------
        data : list
            Merged data.
        """
        return self.fid.merge(data, **kwargs)

    def search_date(self,
                    timestamp,
                    search_date_fmt="%Y%m%d*",
                    date_field="date",
                    date_field_fmt="%Y%m%d",
                    return_date=False,
                    **fmt_kwargs):
        """
        Search files for given date.

        Parameters
        ----------
        timestamp : datetime
            Search date.
        search_date_fmt : str, optional
            Search date string format used during file search (default: %Y%m%d*).
        date_field : str, optional
            Date field name (default: "date")
        date_field_format : str, optional
            Date field string format (default: %Y%m%d).
        return_date : bool, optional
            Return date parsed from filename (default: False).

        Returns
        -------
        filenames : list of str
            Filenames.
        dates : list of datetime
            Parsed date of filename (only returned if return_date=True).
        """
        fn_read_fmt, sf_read_fmt, _, _ = self._fmt(timestamp, **fmt_kwargs)
        fn_read_fmt[date_field] = timestamp.strftime(search_date_fmt)

        fs = FileSearch(self.root_path, self.ft.fn_templ, self.ft.sf_templ)
        def key_func(x): return self._parse_date(x, date_field, date_field_fmt)
        filenames = sorted(fs.search(fn_read_fmt, sf_read_fmt), key=key_func)

        if return_date:
            dates = []
            for filename in filenames:
                dates.append(
                    self._parse_date(filename, date_field, date_field_fmt))
            return filenames, dates
        else:
            return filenames

    def search_period(
        self,
        dt_start,
        dt_end,
        dt_delta=timedelta(days=1),
        search_date_fmt="%Y%m%d*",
        date_field="date",
        date_field_fmt="%Y%m%d",
        end_inclusive=True,
        **fmt_kwargs
    ):
        """
        Search files for time period.

        Parameters
        ----------
        dt_start : datetime
            Start datetime.
        dt_end : datetime
            End datetime.
        dt_delta : timedelta, optional
            Time delta used to jump through search date.
        search_fmt : str, optional
            Search date string format used during file search (default: %Y%m%d*).
        date_field : str, optional
            Date field name (default: "date").
        date_field_fmt : str, optional
            Date field string format (default: %Y%m%d).
        end_inclusive : bool, optional
            Include files from a dt_delta length period beyond dt_end if True
            (default: False).

        Returns
        -------
        filenames : list of str
            Filenames.
        """
        filenames = []

        dt_end = dt_end + dt_delta if end_inclusive else dt_end

        for dt_cur in np.arange(dt_start, dt_end, dt_delta).astype(datetime):
            files, dates = self.search_date(
                dt_cur,
                search_date_fmt=search_date_fmt,
                date_field=date_field,
                date_field_fmt=date_field_fmt,
                return_date=True,
                **fmt_kwargs,
            )
            for f, dt in zip(files, dates):
                if f not in filenames and dt >= dt_start and dt < dt_end:
                    filenames.append(f)

        return filenames

    def read_period(
        self,
        dt_start,
        dt_end,
        dt_delta=timedelta(days=1),
        dt_buffer=timedelta(days=1),
        search_date_fmt="%Y%m%d*",
        date_field="date",
        date_field_fmt="%Y%m%d",
        end_inclusive=True,
        fmt_kwargs={},
        **kwargs,
    ):
        """
        Read data for given interval.

        Parameters
        ----------
        dt_start : datetime
            Start datetime.
        dt_end : datetime
            End datetime.
        dt_delta : timedelta, optional
            Time delta used to jump through search date.
        dt_buffer : timedelta, optional
            Search buffer used to find files which could possibly contain
            data but would be left out because of dt_start.
        search_date_fmt : str, optional
            Search date string format used during file search (default: %Y%m%d*).
        date_field : str, optional
            Date field name (default: "date").
        date_field_fmt : str, optional
            Date field string format (default: %Y%m%d).

        Returns
        -------
        data : dict, numpy.ndarray
            Data stored in file.
        """
        filenames = self.search_period(dt_start - dt_buffer, dt_end, dt_delta,
                                       search_date_fmt, date_field,
                                       date_field_fmt, end_inclusive, **fmt_kwargs)

        data = []

        for filename in filenames:
            self._open(filename)
            d = self.fid.read_period(dt_start, dt_end, **kwargs)
            if d is not None:
                data.append(d)

        if data:
            data = self._merge_data(data)

        return data

class Filenames:
    """
    A class to handle operations on multiple filenames.

    This class provides methods for reading from, writing to, and merging data from multiple files.
    """

    def __init__(self, filenames):
        """
        Initialize Filenames.

        Parameters
        ----------
        filenames : str, Path, or list
            File path(s) to be handled.
        """
        if isinstance(filenames, (str, Path)):
            filenames = [filenames]
        elif not isinstance(filenames, list):
            raise ValueError("filenames must be a string or list of strings.")

        self.filenames = [Path(f) for f in filenames]
        self.cache = {}

    def _read(self, filename, **kwargs):
        """
        Read data from a single file.

        This method should be implemented by subclasses.

        Parameters
        ----------
        filename : Path
            The file to read from.
        **kwargs : dict
            Additional keyword arguments for reading.

        Raises
        ------
        NotImplementedError
            If the method is not implemented in a subclass.
        """
        raise NotImplementedError

    def _merge(self, data, **kwargs):
        """
        Merge multiple data objects.

        This method should be implemented by subclasses.

        Parameters
        ----------
        data : list
            List of data objects to merge.
        **kwargs : dict
            Additional keyword arguments for merging.

        Raises
        ------
        NotImplementedError
            If the method is not implemented in a subclass.
        """
        raise NotImplementedError

    def _write(self, data, filename, **kwargs):
        """
        Write data to a single file.

        This method should be implemented by subclasses.

        Parameters
        ----------
        data : object
            The data to write.
        filename : Path
            The file to write to.
        **kwargs : dict
            Additional keyword arguments for writing.

        Raises
        ------
        NotImplementedError
            If the method is not implemented in a subclass.
        """
        raise NotImplementedError

    def reprocess(self,
                  out_dir,
                  func,
                  parallel=False,
                  print_progress=False,
                  read_kwargs=None,
                  **write_kwargs):
        """
        Reprocess data from all files through `func`, writing the results to `out_dir`.
        Assumes that if any files have the same name, they should be merged.

        Parameters
        ----------
        out_dir : Path
            Directory to write the output files. This will be prepended to the filenames.
        func : function
            The function to apply to the data before writing out.
        parallel : bool, optional
            Whether to process the data in parallel (default: False).
        **kwargs : dict
            Additional keyword arguments for writing.
        """
        read_kwargs = read_kwargs or {}
        if parallel:
            read_ = delayed(self._read)
            getattr_ = delayed(getattr)
            func_ = delayed(func)
            merge_ = delayed(self._merge)
        else:
            read_ = self._read
            getattr_ = getattr
            func_ = func
            merge_ = self._merge

        filenames = self.filenames

        name_to_paths = defaultdict(list)
        for path in filenames:
            name_to_paths[path.name].append(path)

        out_filenames = [out_dir / name for name in name_to_paths]
        out_path_groups = list(name_to_paths.values())

        if print_progress:
            out_path_groups = tqdm(out_path_groups)
            out_path_groups.set_description("Opening files...")

        data = [merge_([func_(read_(f, **read_kwargs)) for f in paths])
                for paths in out_path_groups]

        self.filenames = out_filenames

        self.write(data, parallel=parallel, print_progress=print_progress, **write_kwargs)

    def write(self, data, parallel=False, print_progress=False, **kwargs):
        """
        Write data to file.

        If there's only one filename in `self.filenames`, write provided data to that file.
        If there is more than one filename, write each element of the provided data list
        to the corresponding filename.

        Parameters
        ----------
        data :  list of objects
            The data to write. Should be a list with the same length as self.filenames,
            where each element is the data to be written to the corresponding filename.
        """
        if len(self.filenames) == 1 and not isinstance(data, list):
            data = [data]


        if len(data) == len(self.filenames):
            if parallel:
                write_ = delayed(self._write)
                writers = [write_(d, f, **kwargs) for d, f in zip(data, self.filenames)]
                if print_progress:
                    with TqdmCallback(desc="Writing cells to disk...", total=len(writers)):
                        compute(writers, scheduler="processes")
                else:
                    compute(writers, scheduler="processes")

            else:
                if print_progress:
                    data = tqdm(data)
                    data.set_description("Writing cells to disk...")
                for d, f in zip(data, self.filenames):
                    self._write(d, f, **kwargs)
        else:
            # Special case when the data object meant to be written to a single filename is a list
            raise ValueError("Number of data objects must match number of filenames.")

        return

    def read(self, parallel=False, closer_attr=None, **kwargs):
        """
        Read all data from files.

        Returns
        -------
        object
            Merged data from all files.
        """
        if parallel:
            read_ = delayed(self._read)
            getattr_ = delayed(getattr)
        else:
            read_ = self._read
            getattr_ = getattr

        data = [read_(f, **kwargs) for f in self.filenames]
        if closer_attr is not None:
            closers = [getattr_(d, closer_attr) for d in data if d is not None]

        if parallel:
            data = compute(data, scheduler="processes")[0]
            if closer_attr is not None:
                closers = compute(closers)[0]

        data = self.merge(data)

        if closer_attr is not None:
            return data, closers

        return data

    def iter_read(self, print_progress=False, **kwargs):
        """
        Iterate over all files and yield data.

        Yields
        ------
        object
            Data read from each file.
        """
        if print_progress:
            filenames = tqdm(self.filenames)
        else:
            filenames = self.filenames

        size = 0
        for filename in filenames:
            if print_progress:
                filenames.set_description(f"Opening {Path(filename).name}, total {size} bytes...")
            data = self._read(filename, **kwargs)
            size += self._nbytes(data)
            yield data

    def iter_read_nbytes(self, max_nbytes, print_progress=False, **kwargs):
        """
        Iterate over all files and yield data until the specified number of bytes is reached.
        If `_read` returns dask objects, they are computed (in parallel) before merging the data.
        """
        size = 0
        data_list = []
        for data in self.iter_read(print_progress, **kwargs):
            data_size = self._nbytes(data)
            size += data_size
            if size > max_nbytes and size > data_size:
                if print_progress:
                    print(f"Opened {size} bytes, reading and merging data...")

                    with TqdmCallback(desc="Reading data..."):
                        out_data = compute(*[d for d in data_list],
                                           scheduler="processes")
                    print("Merging data...")
                    out_data = self.merge(out_data)
                else:
                    out_data = self.merge(compute(*[d for d in data_list],
                                                scheduler="processes"))
                yield out_data
                size = data_size
                data_list = [data]
            else:
                data_list.append(data)
        if data_list:
            if print_progress:
                print("All source files opened, reading and merging remaining data...")
            yield self.merge(compute(*[d for d in data_list], scheduler="processes"))

    @staticmethod
    def _nbytes(data):
        """
        Returns size of data object in bytes.
        """
        raise NotImplementedError

    def merge(self, data):
        """
        Merge data from multiple data objects.

        Parameters
        ----------
        data : list
            List of data objects.

        Returns
        -------
        object
            Merged data, or None if the input list is empty.
        """
        if len(data) > 1:
            data = self._merge(data)
        elif len(data) == 1:
            data = data[0]
        else:
            data = None

        return data

    def close(self):
        """
        Close file(s).

        This method can be overridden in subclasses if necessary.
        """
        pass

    @staticmethod
    def _multi_file_closer(closers):
        for closer in closers:
            closer()
        return


class CsvFile(Filenames):
    """
    Read and write single CSV file.
    """

    def __init__(self, filename, mode="r"):
        """
        Initialize Csv

        Parameters
        ----------
        filename : str
            Filename
        mode : str, optional
            File opening mode.
        """
        self.mode = mode
        super().__init__(filename)

    def header2dtype(self, header):
        """
        Convert header string to dtype info.

        Parameters
        ----------
        header : str
            Header string with dtype info.

        Returns
        -------
        dtype : numpy.dtype
            Data type.
        """
        dtype_list = []
        for substr in header.split("(")[2:]:
            d = []
            for substr2 in substr.split(","):
                if substr2.endswith(")"):
                    substr2 = substr2[:-1]
                if substr2.endswith("\n"):
                    substr2 = substr2[:-4]
                substr2 = substr2.strip()
                substr2 = substr2.strip("'")
                if substr2 == "":
                    continue
                d.append(substr2)
            dtype_list.append(tuple(d))

        return np.dtype(dtype_list)

    def _read(self, filename):
        """
        Read data from CSV file.

        Parameters
        ----------
        filename : str
            Filename.
        """
        with open(filename) as fid:
            header = fid.readline()
            dtype = self.header2dtype(header)
            data = np.loadtxt(fid, dtype)

        return data

    def read_period(self, dt_start, dt_end):
        """
        Read subset data from CSV file for given interval.

        Parameters
        ----------
        interval : (datetime, datetime)
            Time interval to extract data.

        Returns
        -------
        data : numpy.ndarray
            Data.
        """
        data = self.read()

        subset = (data["date"] >= np.datetime64(dt_start)) & (
            data["date"] <= np.datetime64(dt_end))

        if np.sum(subset) > 0:
            data = data[subset]
        else:
            data = None

        return data

    def _write(self, data, filename):
        """
        Write data to CSV file.

        Parameters
        ----------
        data : numpy.ndarray
            Data.
        """
        header = data.dtype.__repr__()
        np.savetxt(filename, data, fmt="%s", header=header)

    @staticmethod
    def _merge(data):
        """
        Merge data.

        Parameters
        ----------
        data : list of numpy.ndarray
            List of data.

        Returns
        -------
        data : numpy.ndarray
            Merged data.
        """
        return np.hstack(data)


class CsvFiles(ChronFiles):
    """
    Write CSV files.
    """

    def __init__(self, root_path):
        """
        Initialize CvsFileRW.

        Parameters
        ----------
        root_path : str
            Root path.
        """
        fn_templ = "prefix_{date}_{now}_postfix.csv"
        sf_templ = {"Y": "{year}", "M": "{month}"}

        super().__init__(root_path, CsvFile, fn_templ, sf_templ=sf_templ)

    def _fmt(self, timestamp):
        """
        Definition of filename and subfolder format.

        Parameters
        ----------
        timestamp : datetime
            Time stamp.

        Returns
        -------
        fn_fmt : dict
            Filename format.
        sf_fmt : dict
            Subfolder format.
        """
        fn_read_fmt = {"date": timestamp.strftime("%Y%m%d_%H%M%S"), "now": "*"}

        sf_read_fmt = {
            "Y": {
                "year": timestamp.strftime("%Y")
            },
            "M": {
                "month": timestamp.strftime("%m")
            },
        }

        fn_write_fmt = {
            "date": timestamp.strftime("%Y%m%d_%H%M%S"),
            "now": datetime.now().strftime("%Y%m%d_%H%M%S"),
        }

        sf_write_fmt = sf_read_fmt

        return fn_read_fmt, sf_read_fmt, fn_write_fmt, sf_write_fmt



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Test ASCAT Level 1 reader.
"""

import sys
import pytest
import unittest
from datetime import datetime

import numpy as np
import numpy.testing as nptest

from ascat.eumetsat.level1 import AscatL1bFile
from ascat.eumetsat.level1 import AscatL1bNcFileList
from ascat.eumetsat.level1 import AscatL1bEpsFileList
from ascat.eumetsat.level1 import AscatL1bBufrFileList
from ascat.eumetsat.level1 import AscatL1bHdf5FileList

from get_path import get_testdata_path

float32_nan = -999999.


@pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")
class Test_AscatL1bFile(unittest.TestCase):

    def setUp(self):
        data_path = get_testdata_path(
        ) / "eumetsat" / "ASCAT_generic_reader_data"

        name_b = data_path / "bufr" / "M02-ASCA-ASCSZR1B0200-NA-9.1-20100609013900.000000000Z-20130824233100-1280350.bfr"
        name_e = data_path / "eps_nat" / "ASCA_SZR_1B_M02_20100609013900Z_20100609032058Z_R_O_20130824233100Z.nat"
        name_n = data_path / "nc" / "W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,METOPA+ASCAT_C_EUMP_20100609013900_18872_eps_o_125_l1.nc"

        name_e11 = data_path / "eps_nat" / "ASCA_SZR_1B_M02_20071212071500Z_20071212085659Z_R_O_20081225063118Z.nat"
        name_e_szf = data_path / "eps_nat" / "ASCA_SZF_1B_M01_20180611041800Z_20180611055959Z_N_O_20180611050637Z.nat"
        name_h = data_path / "hdf5" / "ASCA_SZF_1B_M01_20180611041800Z_20180611055959Z_N_O_20180611050637Z.h5"

        self.bufr = AscatL1bFile(name_b)
        self.nc = AscatL1bFile(name_n)

        self.eps = AscatL1bFile(name_e)
        self.eps_fmv11 = AscatL1bFile(name_e11)

        self.eps_szf = AscatL1bFile(name_e_szf)
        self.h5_szf = AscatL1bFile(name_h)

    def test_szx_read(self):
        """
        Test SZX data in all data formats (BUFR, EPS Native, NetCDF).
        """
        bufr_data, metadata = self.bufr.read()
        eps_data, metadata = self.eps.read()
        nc_data, metadata = self.nc.read()

        for f in ["lat", "lon"]:
            nptest.assert_allclose(bufr_data[f], eps_data[f], atol=1e-2)
            nptest.assert_allclose(eps_data[f], nc_data[f], atol=1e-2)
            nptest.assert_allclose(nc_data[f], bufr_data[f], atol=1e-2)

        matching = [
            "sig", "inc", "azi", "kp", "kp_quality", "swath_indicator",
            "f_usable", "f_land", "sat_id", "line_num", "node_num"
        ]

        # BUFR contain less accurate data so we only compare 0.1 accuracy.
        for field in matching:

            not_valid_nc = ((nc_data[field] == float32_nan) |
                            (np.isnan(nc_data[field])))
            not_valid_bufr = ((bufr_data[field] == float32_nan) |
                              (np.isnan(bufr_data[field])))
            not_valid_eps = ((eps_data[field] == float32_nan) |
                             (np.isnan(eps_data[field])))

            nptest.assert_allclose(
                nc_data[field][~(not_valid_nc | not_valid_bufr)],
                bufr_data[field][~(not_valid_nc | not_valid_bufr)],
                atol=0.1)

            nptest.assert_allclose(
                eps_data[field][~(not_valid_eps | not_valid_nc)],
                nc_data[field][~(not_valid_eps | not_valid_nc)],
                atol=0.1)

    def test_read_szf(self):
        """
        Test read SZF formats.
        """
        eps_data, eps_metadata = self.eps_szf.read()
        h5_data, h5_metadata = self.h5_szf.read()

        eps_fields = [
            "sig", "inc", "as_des_pass", "land_frac", "swath_indicator"
        ]

        hdf5_fields = [
            "sigma0_full", "inc_angle_full", "as_des_pass", "land_frac",
            "swath_indicator"
        ]

        for antenna in ["lf-vv", "lm-vv", "la-vv", "rf-vv", "rm-vv", "ra-vv"]:
            for coord in ["lon", "lat"]:
                nptest.assert_allclose(
                    eps_data[antenna][coord],
                    h5_data[antenna][coord],
                    atol=1e-4)

            for eps_field, hdf5_field in zip(eps_fields, hdf5_fields):
                nptest.assert_allclose(
                    eps_data[antenna][eps_field],
                    h5_data[antenna][hdf5_field],
                    atol=0.1)

    def test_szx_eps(self):
        """
        Test read SZX EPS.
        """
        self.reader, metadata = self.eps.read(generic=True)

        lat_should = np.array([
            68.91681, 69.005196, 69.09337, 69.18132, 69.26905, 69.35655,
            69.443825, 69.53087, 69.61768, 69.704254, 69.79059, 69.87668,
            69.962524, 70.04812, 70.13346, 70.218544, 70.303375, 70.38794,
            70.47224, 70.55626, 70.640015, 70.723495, 70.806694, 70.88961,
            70.97224
        ])

        lon_should = np.array([
            168.80144, 168.60977, 168.41656, 168.22179, 168.02544, 167.82748,
            167.62794, 167.42676, 167.22394, 167.01947, 166.81332, 166.60548,
            166.39592, 166.18465, 165.97163, 165.75685, 165.5403, 165.32195,
            165.10178, 164.87979, 164.65594, 164.43024, 164.20264, 163.97314,
            163.74171
        ])

        sig_should = np.array([
            -13.510671, -13.421737, -13.872492, -14.351357, -14.395881,
            -14.382635, -14.860762, -16.108913, -17.354418, -18.86383,
            -18.793966, -18.631758, -18.46626, -18.71435, -19.150038,
            -19.315845, -19.79865, -19.845669, -19.892258, -20.138796,
            -20.151554, -20.154343, -20.165552, -20.013523, -19.238102
        ])

        kp_should = np.array([
            0.0307, 0.032, 0.051, 0.0696, 0.0703, 0.0584, 0.045, 0.0464,
            0.0615, 0.0477, 0.0323, 0.04, 0.0346, 0.0369, 0.0378, 0.0397,
            0.0341, 0.0399, 0.0418, 0.0408, 0.0421, 0.0347, 0.0424, 0.0451,
            0.0523
        ])

        t_should = np.array([
            "2010-06-09T01:39:00.000", "2010-06-09T01:39:00.000",
            "2010-06-09T01:39:00.000", "2010-06-09T01:39:00.000",
            "2010-06-09T01:39:00.000", "2010-06-09T01:39:00.000",
            "2010-06-09T01:39:00.000", "2010-06-09T01:39:01.874",
            "2010-06-09T01:39:01.874", "2010-06-09T01:39:01.874"
        ],
                            dtype="datetime64[ms]")

        nptest.assert_allclose(self.reader["lat"][:25], lat_should, atol=1e-5)
        nptest.assert_allclose(self.reader["lon"][:25], lon_should, atol=1e-5)
        nptest.assert_allclose(
            self.reader["sig"][:25, 0], sig_should, atol=1e-5)
        nptest.assert_allclose(self.reader["kp"][:25, 0], kp_should, atol=1e-5)
        nptest.assert_equal(self.reader["time"][75:85], t_should)

    def test_szx_eps_fmv11(self):
        """
        Test read SZX EPS format version 11.
        """
        self.reader, metadata = self.eps_fmv11.read()

        lat_should = np.array([
            61.849445, 61.916786, 61.983864, 62.050674, 62.11722, 62.183495,
            62.2495, 62.315228, 62.380684, 62.44586, 62.51076, 62.57538,
            62.63971, 62.70376, 62.76752, 62.830994, 62.894173, 62.95706,
            63.019653, 63.081947, 63.143944, 63.205635, 63.267025, 63.32811,
            63.388885
        ])

        lon_should = np.array([
            69.18133, 68.991295, 68.80043, 68.60872, 68.41617, 68.22277,
            68.028534, 67.833435, 67.63749, 67.44069, 67.243034, 67.04452,
            66.84513, 66.64489, 66.44378, 66.2418, 66.03894, 65.83522,
            65.630615, 65.42514, 65.21878, 65.01154, 64.80342, 64.59441,
            64.38452
        ])

        sig_should = np.array([
            -15.008125, -14.547356, -15.067405, -15.340037, -15.381483,
            -15.085848, -14.620477, -14.200545, -13.873865, -13.29581,
            -12.962119, -12.909232, -12.990307, -13.076723, -13.039384,
            -13.010556, -13.238036, -13.045113, -12.981088, -13.003889,
            -14.009461, -14.633162, -14.706434, -14.042056, -13.7074
        ])

        kp_should = np.array([
            0.052, 0.0417, 0.0462, 0.0264, 0.0308, 0.0296, 0.0363, 0.0348,
            0.0377, 0.036, 0.0329, 0.0258, 0.0296, 0.0245, 0.0275, 0.0309,
            0.035, 0.0325, 0.0288, 0.0292, 0.0431, 0.0363, 0.0435, 0.0282,
            0.0309
        ])

        t_should = np.array([
            "2007-12-12T07:15:00.000", "2007-12-12T07:15:00.000",
            "2007-12-12T07:15:00.000", "2007-12-12T07:15:00.000",
            "2007-12-12T07:15:00.000", "2007-12-12T07:15:00.000",
            "2007-12-12T07:15:00.000", "2007-12-12T07:15:01.876",
            "2007-12-12T07:15:01.876", "2007-12-12T07:15:01.876"
        ],
                            dtype="datetime64[ms]")

        nptest.assert_allclose(self.reader["lat"][:25], lat_should, atol=1e-5)
        nptest.assert_allclose(self.reader["lon"][:25], lon_should, atol=1e-5)
        nptest.assert_allclose(
            self.reader["sig"][:25, 0], sig_should, atol=1e-5)
        nptest.assert_allclose(self.reader["kp"][:25, 0], kp_should, atol=1e-5)
        nptest.assert_equal(self.reader["time"][75:85], t_should)

    def test_szf_eps(self):
        """
        Test read SZF EPS format.
        """
        self.reader, metadata = self.eps_szf.read()

        lat_should = np.array([
            64.45502, 64.42318, 64.39127, 64.35929, 64.32724, 64.29512,
            64.262924, 64.23065, 64.19831, 64.16589, 64.1334, 64.10083,
            64.06819, 64.03547, 64.00268, 63.969807, 63.936855, 63.903828,
            63.870724, 63.837536, 63.804276, 63.77093, 63.737507, 63.704002,
            63.67042
        ])

        lon_should = np.array([
            103.29956, 103.32185, 103.34413, 103.36641, 103.38869, 103.41095,
            103.43322, 103.45548, 103.47774, 103.49999, 103.52224, 103.54449,
            103.566734, 103.588974, 103.61121, 103.63345, 103.655685,
            103.67792, 103.70014, 103.722374, 103.7446, 103.76682, 103.78904,
            103.811264, 103.83348
        ])

        sig_should = np.array([
            -9.713457, -8.768949, -9.294478, -7.449275, -8.939872, -7.893198,
            -8.570546, -8.934691, -7.851117, -7.782818, -8.33993, -7.539894,
            -7.833797, -8.465893, -8.244121, -7.59996, -8.976448, -9.36595,
            -10.800382, -8.289896, -9.127579, -9.410345, -7.238986, -8.335969,
            -7.897769
        ])

        t_should = np.array([
            "2018-06-11T04:18:00.630", "2018-06-11T04:18:00.630",
            "2018-06-11T04:18:01.479", "2018-06-11T04:18:01.479",
            "2018-06-11T04:18:01.479", "2018-06-11T04:18:01.479",
            "2018-06-11T04:18:01.479", "2018-06-11T04:18:01.479",
            "2018-06-11T04:18:01.479", "2018-06-11T04:18:01.479"
        ],
                            dtype="datetime64[ms]")

        nptest.assert_allclose(
            self.reader["lf-vv"]["lat"][:25], lat_should, atol=1e-5)
        nptest.assert_allclose(
            self.reader["lf-vv"]["lon"][:25], lon_should, atol=1e-5)
        nptest.assert_allclose(
            self.reader["lf-vv"]["sig"][:25], sig_should, atol=1e-5)
        nptest.assert_equal(self.reader["lf-vv"]["time"][190:200], t_should)


@pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")
class Test_AscatL1bFileList(unittest.TestCase):
    """
    Test read AscatL1bFileList in various formats.
    """

    def setUp(self):
        """
        Setup test data.
        """
        root_path = get_testdata_path(
        ) / "eumetsat" / "ASCAT_generic_reader_data"

        path = root_path / "bufr"
        sat = "a"
        product = "szr"
        self.bufr_szr = AscatL1bBufrFileList(path, sat, product)

        path = root_path / "nc"
        sat = "a"
        product = "szr"
        self.nc_szr = AscatL1bNcFileList(path, sat, product)

        path = root_path / "eps_nat"
        sat = "a"
        product = "szr"
        self.eps_szr = AscatL1bEpsFileList(path, sat, product)

        sat = "b"
        product = "szf"
        path = root_path / "eps_nat"
        self.eps_szf = AscatL1bEpsFileList(path, sat, product)

        sat = "b"
        product = "szf"
        path = root_path / "hdf5"
        self.hdf5_szf = AscatL1bHdf5FileList(path, sat, product)

    def test_szr_read_date(self):
        """
        Test read date for SZR formats.
        """
        dt = datetime(2010, 6, 9, 1, 39, 0)

        bufr_data, metadata = self.bufr_szr.read(dt)
        nc_data, metadata = self.nc_szr.read(dt)
        eps_data, metadata = self.eps_szr.read(dt)

        for f in ["lat", "lon"]:
            nptest.assert_allclose(bufr_data[f], eps_data[f], atol=1e-2)
            nptest.assert_allclose(eps_data[f], nc_data[f], atol=1e-2)
            nptest.assert_allclose(nc_data[f], bufr_data[f], atol=1e-2)

        matching = [
            "sig", "inc", "azi", "kp", "kp_quality", "swath_indicator",
            "f_usable", "f_land", "sat_id", "line_num", "node_num"
        ]

        # BUFR contain less accurate data so we only compare 0.1 accuracy.
        for field in matching:

            not_valid_nc = ((nc_data[field] == float32_nan) |
                            (np.isnan(nc_data[field])))
            not_valid_bufr = ((bufr_data[field] == float32_nan) |
                              (np.isnan(bufr_data[field])))
            not_valid_eps = ((eps_data[field] == float32_nan) |
                             (np.isnan(eps_data[field])))

            nptest.assert_allclose(
                nc_data[field][~(not_valid_nc | not_valid_bufr)],
                bufr_data[field][~(not_valid_nc | not_valid_bufr)],
                atol=0.1)

            nptest.assert_allclose(
                eps_data[field][~(not_valid_eps | not_valid_nc)],
                nc_data[field][~(not_valid_eps | not_valid_nc)],
                atol=0.1)

    def test_szr_read_period(self):
        """
        Test read period for SZR formats.
        """
        dt_start = datetime(2010, 6, 9, 1, 39, 0)
        dt_end = datetime(2010, 6, 9, 2, 0, 0)

        date_field_fmt = "%Y%m%d%H%M%S"

        bufr_data, bufr_metadata = self.bufr_szr.read_period(
            dt_start, dt_end, date_field_fmt=date_field_fmt)
        nc_data, nc_metadata = self.nc_szr.read_period(
            dt_start, dt_end, date_field_fmt=date_field_fmt)
        eps_data, eps_metadata = self.eps_szr.read_period(
            dt_start, dt_end, date_field_fmt=date_field_fmt)

        for f in ["lat", "lon"]:
            nptest.assert_allclose(bufr_data[f], eps_data[f], atol=1e-2)
            nptest.assert_allclose(eps_data[f], nc_data[f], atol=1e-2)
            nptest.assert_allclose(nc_data[f], bufr_data[f], atol=1e-2)

        matching = [
            "kp_quality", "swath_indicator", "f_usable", "f_land", "sat_id",
            "line_num", "node_num"
        ]

        for field in matching:
            nptest.assert_allclose(nc_data[field], bufr_data[field], atol=0.1)
            nptest.assert_allclose(eps_data[field], nc_data[field], atol=0.1)

        matching = ["sig", "inc", "azi", "kp"]

        # BUFR contain less accurate data so we only compare 0.1 accuracy.
        for field in matching:
            nan_mask = (nc_data[field] == float32_nan)
            eps_data[field][nan_mask] = float32_nan
            bufr_data[field][nan_mask] = float32_nan
            valid = ((eps_data[field] > -25))

            nptest.assert_allclose(
                bufr_data[field][valid], eps_data[field][valid], atol=0.1)

    def test_szf_read_date(self):
        """
        Test read date for SZF formats.
        """
        dt = datetime(2018, 6, 11, 4, 18, 0)

        eps_data, metadata = self.eps_szf.read(dt)
        hdf5_data, metadata = self.hdf5_szf.read(dt)

        eps_fields = [
            "sig", "inc", "as_des_pass", "land_frac", "swath_indicator"
        ]

        hdf5_fields = [
            "sigma0_full", "inc_angle_full", "as_des_pass", "land_frac",
            "swath_indicator"
        ]

        for antenna in ["lf-vv", "lm-vv", "la-vv", "rf-vv", "rm-vv", "ra-vv"]:
            for coord in ["lon", "lat"]:
                nptest.assert_allclose(
                    eps_data[antenna][coord],
                    hdf5_data[antenna][coord],
                    atol=1e-4)

            for eps_field, hdf5_field in zip(eps_fields, hdf5_fields):
                nptest.assert_allclose(
                    eps_data[antenna][eps_field],
                    hdf5_data[antenna][hdf5_field],
                    atol=0.1)

    def test_szf_read_period(self):
        """
        Test read period for SZF formats.
        """
        dt_start = datetime(2018, 6, 11, 4, 18, 0)
        dt_end = datetime(2018, 6, 11, 4, 19, 0)

        date_field_fmt = "%Y%m%d%H%M%S"
        eps_data, eps_metadata = self.eps_szf.read_period(
            dt_start, dt_end, date_field_fmt=date_field_fmt)
        hdf5_data, hdf5_metadata = self.hdf5_szf.read_period(
            dt_start, dt_end, date_field_fmt=date_field_fmt)

        for antenna in ["lf-vv", "lm-vv", "la-vv", "rf-vv", "rm-vv", "ra-vv"]:
            for coord in ["lon", "lat"]:
                nptest.assert_allclose(
                    eps_data[antenna][coord],
                    hdf5_data[antenna][coord],
                    atol=1e-4)

            matching = [
                "sig", "inc", "azi", "sat_id", "as_des_pass", "land_frac",
                "f_usable", "f_land", "beam_number", "swath_indicator"
            ]

            for field in matching:
                nptest.assert_allclose(
                    eps_data[antenna][coord],
                    hdf5_data[antenna][coord],
                    atol=0.1)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

from typing import Union, Sequence

import numpy as np
import xarray as xr


def check_orthomulti_ts(ds):
    # Assumptions:
    # - two dimensions [DONE]
    # - single variable with only the sample dimension (e.g. time) [TODO]
    # - data variables have sample and instance dimension [TODO]
    # - data variables have ALL instance dimension coordinates listed as coordinates [TODO]
    if len(ds.dims) == 2:
        for v in ds.variables:
            if "cf_role" in ds[v].attrs and ds[v].attrs["cf_role"] == "timeseries_id":
                return True
    return False


def cf_array_type(ds):
    if ds.attrs.get("featureType") == "point":
        return "point"
    for v in ds.variables:
        if "instance_dimension" in ds[v].attrs:
            return "indexed"
        if "sample_dimension" in ds[v].attrs:
            return "contiguous"
    if check_orthomulti_ts(ds):
        return "orthomulti_ts"
    raise ValueError("Array type could not be determined.")


def cf_array_class(ds, array_type, **kwargs):
    if array_type == "point":
        return TimeseriesPointArray(ds, **kwargs)
    if array_type == "indexed":
        return RaggedArray(ds, **kwargs)
    if array_type == "contiguous":
        return RaggedArray(ds, **kwargs)
    if array_type == "orthomulti_ts":
        return OrthoMultiTimeseriesArray(ds, **kwargs)
    raise ValueError(f"Array type '{array_type}' not recognized."
                     "Should be one of 'point', 'indexed', 'contiguous', 'orthomulti_ts'.")


def point_to_indexed(
    ds: xr.Dataset,
    sample_dim: str,
    instance_dim: str,
    timeseries_id: str,
    index_var: str = "locationIndex",
    instance_vars: Union[Sequence[str], None] = None,
    coord_vars: Union[Sequence[str], None] = None,
) -> xr.Dataset:
    coord_vars = coord_vars or []
    instance_vars = instance_vars or []
    instance_vars = [timeseries_id] + instance_vars

    _, unique_index_1d, instanceIndex = np.unique(
        ds[timeseries_id], return_index=True, return_inverse=True
    )
    ds[index_var] = (sample_dim, instanceIndex, {"instance_dimension": instance_dim})

    for var in instance_vars:
        if var in ds:
            ds = ds.assign(
                {var: (instance_dim, ds[var][unique_index_1d].data, ds[var].attrs)}
            )
            if var in coord_vars:
                ds = ds.set_coords(var)
    ds = ds.assign_attrs({"featureType": "timeSeries"})
    return ds


def point_to_contiguous(
    ds: xr.Dataset,
    sample_dim: str,
    instance_dim: str,
    timeseries_id: str,
    count_var: str = "row_size",
    instance_vars: Union[Sequence[str], None] = None,
    coord_vars: Union[Sequence[str], None] = None,
    sort_vars: Union[Sequence[str], None] = None,
) -> xr.Dataset:
    coord_vars = coord_vars or []
    sort_vars = sort_vars or []
    instance_vars = instance_vars or []
    instance_vars = [timeseries_id] + instance_vars


    ds = ds.sortby([timeseries_id, *sort_vars])
    _, unique_index_1d, row_size = np.unique(
        ds[timeseries_id], return_index=True, return_counts=True
    )

    ds[count_var] = ("locations", row_size, {"sample_dimension": sample_dim})

    for var in instance_vars:
        if var in ds:
            encoding = ds[var].encoding
            ds[var] = (instance_dim, ds[var][unique_index_1d].data, ds[var].attrs)
            ds[var].encoding = encoding
            if var in coord_vars:
                ds = ds.set_coords(var)
    ds = ds.assign_attrs({"featureType": "timeSeries"})
    return ds


def contiguous_to_indexed(
    ds: xr.Dataset,
    sample_dim: str,
    instance_dim: str,
    count_var: str,
    index_var: str,
) -> xr.Dataset:
    """
    Convert a contiguous ragged array dataset to an indexed ragged array dataset.
    """
    row_size = np.where(ds[count_var].data > 0, ds[count_var].data, 0)

    locationIndex = np.repeat(np.arange(row_size.size), row_size)

    ds = ds.assign(
        {
            index_var: (
                sample_dim,
                locationIndex,
                {"instance_dimension": instance_dim},
            )
        }
    ).drop_vars([count_var])

    # put locationIndex as first var
    ds = ds[[index_var] + [var for var in ds.variables if var != index_var]]

    return ds


def indexed_to_contiguous(
    ds: xr.Dataset,
    sample_dim: str,
    instance_dim: str,
    count_var: str,
    index_var: str,
    sort_vars: Union[Sequence[str], None] = None,
) -> xr.Dataset:
    """
    Convert an indexed ragged array dataset to a contiguous ragged array dataset
    """
    sort_vars = sort_vars or []

    ds = ds.sortby([index_var, *sort_vars])
    idxs, sizes = np.unique(ds[index_var], return_counts=True)

    row_size = np.zeros_like(ds[instance_dim].data)
    row_size[idxs] = sizes
    ds = ds.assign(
        {count_var: (instance_dim, row_size, {"sample_dimension": sample_dim})}
    ).drop_vars([index_var])

    return ds


def indexed_to_point(
    ds: xr.Dataset, sample_dim: str, instance_dim: str, index_var: str
):
    instance_vars = [var for var in ds.variables if instance_dim in ds[var].dims]
    for instance_var in instance_vars:
        ds = ds.assign(
            {
                instance_var: (
                    sample_dim,
                    ds[instance_var][ds[index_var]].data,
                    ds[instance_var].attrs,
                )
            }
        )
    ds = ds.drop_vars([index_var]).assign_attrs({"featureType": "point"})
    return ds


def contiguous_to_point(
    ds: xr.Dataset,
    sample_dim: str,
    instance_dim: str,
    count_var: str,
):
    """Convert a contiguous ragged array dataset to a Point Array.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset.
    sample_dim : str
        Name of the sample dimension.
    instance_dim : str
        Name of the instance dimension.
    count_var : str
        Name of the count variable.

    Returns
    -------
    xarray.Dataset
        Dataset with only the time series variables.
    """
    row_size = ds[count_var].values
    ds = ds.drop_vars(count_var)
    instance_vars = [var for var in ds.variables if instance_dim in ds[var].dims]
    for instance_var in instance_vars:
        ds = ds.assign(
            {
                instance_var: (
                    sample_dim,
                    np.repeat(ds[instance_var].values, row_size),
                    ds[instance_var].attrs,
                )
            }
        )
    ds = ds.assign_attrs({"featureType": "point"})
    return ds


class CFDiscreteGeom:
    def __init__(
        self,
        xarray_obj: xr.Dataset,
        coord_vars: Union[Sequence[str], None] = None,
        instance_vars: Union[Sequence[str], None] = None,
        contiguous_sort_vars: Union[Sequence[str], None] = None,
    ):
        """
        Parameters
        ----------
        xarray_obj : xarray.Dataset
            Xarray dataset.
        coord_vars : Sequence[str], optional
            Coordinate variables, by default None.
        instance_vars : Sequence[str], optional
            Instance variables, by default None.
        contiguous_sort_vars : Sequence[str], optional
            Variables that each timeseries should be sorted by in contiguous ragged array format.
        """
        self._data = xarray_obj
        self._coord_vars = coord_vars or [
            "lon",
            "lat",
            "alt",
            "longitude",
            "latitude",
            "altitude",
        ]
        self._instance_vars = instance_vars or [
            "lon",
            "lat",
            "alt",
            "longitude",
            "latitude",
            "altitude",
            "location_description",
        ]

        self._contiguous_sort_vars = contiguous_sort_vars or [
            "time",
        ]

        self._ra_type = None
        self._sample_dimension = None
        self._instance_dimension = None
        self._count_var = None
        self._index_var = None
        self._timeseries_id = None
        self.array_type

    @property
    def array_type():
        raise NotImplementedError


class PointArray(CFDiscreteGeom):
    pass


class TimeseriesPointArray(PointArray):
    """
    Assumptions made beyond basic CF conventions:

    - cf_role="timeseries_id" is used to identify the timeseries ID variable for purposes
        of selecting instances and converting to ragged arrays. If you only have a single
        timeseries there's not much point in using this class.
    """
    @property
    def array_type(self) -> str:
        if self._ra_type is None:
            if self._data.attrs["featureType"] == "point":
                self._ra_type = "point"
                self._sample_dimension = str(list(self._data.dims)[0])
            else:
                raise ValueError(
                    "Dataset is not a point array"
                    "(should have featureType='point' in attributes)."
                )
        return self._ra_type

    @property
    def timeseries_id(self):
        if self._timeseries_id is not None:
            return self._timeseries_id
        for v in self._data.variables:
            if cf_role := self._data[v].attrs.get("cf_role"):
                if cf_role == "timeseries_id":
                    self._timeseries_id = v
                    return self.timeseries_id
        raise ValueError(
            "Timeseries ID could not be determined from dataset attributes."
        )

    def sel_instances(
        self,
        instance_vals: Union[Sequence[Union[int, str]], np.ndarray, None] = None,
        instance_lookup_vector: Union[np.ndarray, None] = None,
        timeseries_id: str = "location_id",
    ):
        ds = self._data
        return self._select_instances(
            ds,
            self._sample_dimension,
            instance_vals,
            instance_lookup_vector,
            timeseries_id,
        )

    def to_indexed_ragged(
        self,
        instance_dim: str = "locations",
        timeseries_id: str = "location_id",
        index_var: str = "locationIndex",
        instance_vars: Union[Sequence[str], None] = None,
        coord_vars: Union[Sequence[str], None] = None,
    ) -> xr.Dataset:
        return self._point_to_indexed(
            self._data,
            self._sample_dimension,
            instance_dim,
            timeseries_id,
            index_var,
            instance_vars or self._instance_vars,
            coord_vars or self._coord_vars,
        )

    def to_contiguous_ragged(
        self,
        instance_dim: str = "locations",
        timeseries_id: str = "location_id",
        count_var: str = "row_size",
        instance_vars: Union[Sequence[str], None] = None,
        coord_vars: Union[Sequence[str], None] = None,
        sort_vars: Union[Sequence[str], None] = None,
    ) -> xr.Dataset:
        return self._point_to_contiguous(
            self._data,
            self._sample_dimension,
            instance_dim,
            timeseries_id,
            count_var,
            instance_vars or self._instance_vars,
            coord_vars or self._coord_vars,
            sort_vars or self._contiguous_sort_vars,
        )

    def to_orthomulti(
            self,
            instance_dim: str = "locations",
            timeseries_id: str = "location_id",
            count_var: str = "row_size",
            instance_vars: Union[Sequence[str], None] = None,
            coord_vars: Union[Sequence[str], None] = None,
            sort_vars: Union[Sequence[str], None] = None,
    ):
        return self._point_to_orthomulti(
            self._data,
            self._sample_dimension,
            instance_dim,
            timeseries_id,
            count_var,
            instance_vars or self._instance_vars,
            coord_vars or self._coord_vars,
            sort_vars or self._contiguous_sort_vars,
        )

    def resample_to_orthomulti(
            self,
            instance_dim: str = "locations",
            timeseries_id: str = "location_id",
            count_var: str = "row_size",
            instance_vars: Union[Sequence[str], None] = None,
            coord_vars: Union[Sequence[str], None] = None,
            sort_vars: Union[Sequence[str], None] = None,
            vars_to_resample: Union[Sequence[str], None] = None,
            resample_method: callable = np.mean,
            resample_period: str = "1M",
    ):
        return self._resample_point_to_orthomulti(
            self._data,
            self._sample_dimension,
            instance_dim,
            timeseries_id,
            count_var,
            instance_vars or self._instance_vars,
            coord_vars or self._coord_vars,
            sort_vars or self._contiguous_sort_vars,
            vars_to_resample,
            resample_method,
            resample_period,
        )

    def to_point_array(self):
        return self._data

    def set_sample_dimension(self, sample_dim: str):
        if self._sample_dimension != sample_dim:
            self._data = self._data.rename_dims({self._sample_dimension: sample_dim})
            self._sample_dimension = sample_dim
        return self._data

    @staticmethod
    def _select_instances(
        ds: xr.Dataset,
        sample_dim: str,
        instance_vals: Union[Sequence[Union[int, str]], np.ndarray, None] = None,
        instance_lookup_vector: Union[np.ndarray, None] = None,
        timeseries_id: str = "location_id",
    ) -> xr.Dataset:
        if not ds.chunks:
            ds = ds.chunk({sample_dim: -1})
        if instance_vals is None:
            instance_vals = []
        if instance_lookup_vector is not None:
            sample_idx = instance_lookup_vector[ds[timeseries_id]]
            return ds.sel({sample_dim: sample_idx})
        sample_idx = np.isin(ds[timeseries_id], instance_vals)
        return ds.sel({sample_dim: sample_idx})

    @staticmethod
    def _point_to_indexed(
        ds: xr.Dataset,
        sample_dim: str,
        instance_dim: str,
        timeseries_id: str,
        index_var: str = "locationIndex",
        instance_vars: Union[Sequence[str], None] = None,
        coord_vars: Union[Sequence[str], None] = None,
    ) -> xr.Dataset:
        return point_to_indexed(
            ds,
            sample_dim,
            instance_dim,
            timeseries_id,
            index_var,
            instance_vars,
            coord_vars,
        )

    @staticmethod
    def _point_to_contiguous(
        ds: xr.Dataset,
        sample_dim: str,
        instance_dim: str,
        timeseries_id: str,
        count_var: str = "row_size",
        instance_vars: Union[Sequence[str], None] = None,
        coord_vars: Union[Sequence[str], None] = None,
        sort_vars: Union[Sequence[str], None] = None,
    ) -> xr.Dataset:
        return point_to_contiguous(
            ds,
            sample_dim,
            instance_dim,
            timeseries_id,
            count_var,
            instance_vars,
            coord_vars,
            sort_vars,
        )

    @staticmethod
    def _point_to_orthomulti(
        ds: xr.Dataset,
        sample_dim: str,
        instance_dim: str,
        timeseries_id: str,
        count_var: str = "row_size",
        instance_vars: Union[Sequence[str], None] = None,
        coord_vars: Union[Sequence[str], None] = None,
        sort_vars: Union[Sequence[str], None] = None,
    ) -> xr.Dataset:
        """
        At the moment, minimum resolution is 1D
        """
        ds = ds.rename({sample_dim: "time"}).set_xindex("time")
        ds = ds.set_index(event=["time", timeseries_id]).unstack("event")
        for c in ds.coords:
            if "time" in ds[c].dims and c != "time":
                ds[c] = ds[c].max("time", keep_attrs=True)
        ds.attrs.pop("featureType")
        return ds


    @staticmethod
    def _resample_point_to_orthomulti(
        ds: xr.Dataset,
        sample_dim: str,
        instance_dim: str,
        timeseries_id: str,
        count_var: str = "row_size",
        instance_vars: Union[Sequence[str], None] = None,
        coord_vars: Union[Sequence[str], None] = None,
        sort_vars: Union[Sequence[str], None] = None,
        vars_to_resample: Union[Sequence[str], None] = None,
        resample_method: callable = np.mean,
        resample_period: str = "1M",
    ) -> xr.Dataset:
        """
        At the moment, minimum resolution is 1D
        """
        ds = ds.rename({sample_dim: "time"}).set_xindex("time")
        ds = ds.set_index(event=["time", timeseries_id]).unstack("event")
        ds = ds.resample(time=resample_period).apply(resample_method)
        ds.attrs.pop("featureType")
        return ds



class RaggedArray(CFDiscreteGeom):
    @property
    def array_type(self):
        if self._ra_type is not None:
            return self._ra_type

        ds = self._data
        for v in ds.variables:
            if "instance_dimension" in ds[v].attrs:
                self._ra_type = "indexed"
                self._index_var = v
                self._instance_dimension = ds[v].attrs["instance_dimension"]
                self._sample_dimension = str(ds[v].dims[0])
                return self._ra_type

            if "sample_dimension" in ds[v].attrs:
                self._ra_type = "contiguous"
                self._count_var = v
                self._sample_dimension = ds[v].attrs["sample_dimension"]
                if len(ds[v].dims) > 0:
                    self._instance_dimension = ds[v].dims[0]
                return self._ra_type

        raise ValueError("Ragged array type could not be determined.")

    @property
    def timeseries_id(self):
        if self._timeseries_id is not None:
            return self._timeseries_id
        for v in self._data.variables:
            if cf_role := self._data[v].attrs.get("cf_role"):
                if cf_role == "timeseries_id":
                    self._timeseries_id = v
                    return self.timeseries_id
        raise ValueError(
            "Timeseries ID could not be determined from dataset attributes."
        )

    def to_indexed_ragged(
            self,
            index_var: str = "locationIndex"
    ) -> xr.Dataset:
        if self.array_type == "indexed":
            return self._data
        elif self.array_type == "contiguous":
            if self._index_var is None:
                self._index_var = index_var
            return self._contiguous_to_indexed(
                self._data,
                self._sample_dimension,
                self._instance_dimension,
                self._count_var,
                self._index_var,
            )

    def to_contiguous_ragged(
        self,
        count_var: str = "row_size",
        sort_vars: Union[Sequence[str], None] = None
    ) -> xr.Dataset:
        if self.array_type == "contiguous":
            return self._data
        elif self.array_type == "indexed":
            if self._count_var is None:
                self._count_var = count_var
            return self._indexed_to_contiguous(
                self._data,
                self._sample_dimension,
                self._instance_dimension,
                self._count_var,
                self._index_var,
                sort_vars=sort_vars or self._contiguous_sort_vars,
            )

    def to_point_array(self):
        if self.array_type == "indexed":
            return self._indexed_to_point(
                self._data,
                self._sample_dimension,
                self._instance_dimension,
                self._index_var,
            )
        if self.array_type == "contiguous":
            return self._contiguous_to_point(
                self._data,
                self._sample_dimension,
                self._instance_dimension,
                self._count_var,
            )

    def sel_instances(
        self,
        instance_vals: Union[Sequence[Union[int, str]], np.ndarray, None] = None,
        instance_lookup_vector: Union[np.ndarray, None] = None,
    ) -> xr.Dataset:
        if self.array_type == "indexed":
            # convert to point array, select there, convert back\
            ds = self.to_point_array()
            instances = ds.cf_geom.sel_instances(
                instance_vals=instance_vals,
                instance_lookup_vector=instance_lookup_vector,
            )
            return instances.cf_geom.to_indexed_ragged(index_var=self._index_var)

        if self.array_type == "contiguous":
            return self._select_instances_contiguous(
                self._data,
                self._sample_dimension,
                self._instance_dimension,
                self.timeseries_id,
                self._count_var,
                instance_vals=instance_vals,
                instance_lookup_vector=instance_lookup_vector,
            )


    def set_sample_dimension(self, sample_dim: str):
        if self._sample_dimension != sample_dim:
            self._data = self._data.rename_dims({self._sample_dimension: sample_dim})
            if self.array_type == "contiguous":
                self._data[self._count_var].attrs["sample_dimension"] = sample_dim
            self._sample_dimension = sample_dim
        return self._data


    @staticmethod
    def _select_instances_contiguous(
        ds: xr.Dataset,
        sample_dim: str,
        instance_dim: str,
        timeseries_id: str,
        count_var: str,
        instance_vals: Union[Sequence[int], np.ndarray, None] = None,
        instance_lookup_vector: Union[np.ndarray, None] = None,
    ) -> xr.Dataset:
        if instance_vals is None:
            instance_vals = []

        # For contiguous using the lookup vector would be slower, so if we get only that,
        # we'll just turn it into an instance_vals array.
        if len(instance_vals) == 0:
            if instance_lookup_vector is not None and sum(instance_lookup_vector) > 0:
                instance_vals = np.where(instance_lookup_vector)[0]

        def get_single_instance_idxs(ds, instance_val):
            instances_idx = np.where(ds[timeseries_id] == instance_val)[0]
            if len(instances_idx) == 0:
                return None
            instances_idx = int(instances_idx[0])
            sample_start = int(
                ds[count_var].isel({instance_dim: slice(0, instances_idx)}).sum().values
            )
            sample_end = int(
                sample_start + ds[count_var].isel({instance_dim: instances_idx}).values
            )
            return sample_start, sample_end, instances_idx

        def select_single_instance(ds, sample_start, sample_end, instances_idx):
            return ds.isel(
                {
                    sample_dim: slice(sample_start, sample_end),
                    instance_dim: instances_idx,
                }
            )

        def select_several_instances(ds, sample_starts, sample_ends, instances_idxs):
            sample_idxs = np.concatenate(
                [range(start, end)
                 for start, end
                 in zip(sample_starts, sample_ends)
                 if end > start]
            )
            return ds.isel({sample_dim: sample_idxs,
                            instance_dim: np.array(instances_idxs)})


        if len(instance_vals) == 1:
            if get_single_instance_idxs(ds, instance_vals[0]) is None:
                return None
            return select_single_instance(ds, *get_single_instance_idxs(ds, instance_vals[0]))
        else:
            instance_vals = np.unique(instance_vals)
            ds[count_var].load()
            ds[timeseries_id].load()
            results = [get_single_instance_idxs(ds, instance_val)
                       for instance_val in instance_vals]
            results = [r for r in results if r is not None]
            if len(results) == 0:
                return None
            if not ds.chunks:
                ds = ds.chunk({sample_dim: -1})
            return select_several_instances(
                ds,
                *zip(*results)
            )


    @staticmethod
    def _contiguous_to_indexed(
        ds: xr.Dataset,
        sample_dim: str,
        instance_dim: str,
        count_var: str,
        index_var: str,
    ) -> xr.Dataset:
        """
        Convert a contiguous ragged array dataset to an indexed ragged array dataset.
        """
        return contiguous_to_indexed(ds, sample_dim, instance_dim, count_var, index_var)

    @staticmethod
    def _indexed_to_contiguous(
        ds: xr.Dataset,
        sample_dim: str,
        instance_dim: str,
        count_var: str,
        index_var: str,
        sort_vars: Union[Sequence[str], None] = None,
    ) -> xr.Dataset:
        """
        Convert an indexed ragged array dataset to a contiguous ragged array dataset
        """
        return indexed_to_contiguous(
            ds, sample_dim, instance_dim, count_var, index_var, sort_vars
        )

    @staticmethod
    def _indexed_to_point(
        ds: xr.Dataset, sample_dim: str, instance_dim: str, index_var: str
    ):
        return indexed_to_point(ds, sample_dim, instance_dim, index_var)

    @staticmethod
    def _contiguous_to_point(
        ds: xr.Dataset,
        sample_dim: str,
        instance_dim: str,
        count_var: str,
    ):
        """Convert a ragged array dataset to a Point Array.

        Parameters
        ----------
        ds : xarray.Dataset
            Dataset.

        Returns
        -------
        xarray.Dataset
            Dataset with only the time series variables.
        """
        return contiguous_to_point(ds, sample_dim, instance_dim, count_var)


class OrthoMultiTimeseriesArray(CFDiscreteGeom):
    @property
    def array_type(self):
        if check_orthomulti_ts(self._data):
            for v in self._data.variables:
                if "cf_role" in self._data[v].attrs and self._data[v].attrs["cf_role"] == "timeseries_id":
                    self._timeseries_id = v
                    self._instance_dimension = self._data[v].dims[0]
                    break
            return "orthomulti_ts"
        else:
            raise ValueError("Dataset is not an orthomulti timeseries array.")

    def sel_instances(
        self,
        instance_vals: Union[Sequence[Union[int, str]], np.ndarray, None] = None,
        instance_lookup_vector: Union[np.ndarray, None] = None,
    ):
        """
        Select requested timeseries instances from an orthomulti timeseries array dataset.

        Parameters
        ----------
        instance_vals : Union[Sequence[Union[int, str]], np.ndarray], optional
            List of instance values to select, by default None
        instance_lookup_vector : Union[np.ndarray], optional
            Lookup vector for instance values, by default None
        """
        return self._select_instances(
            self._data,
            self._instance_dimension,
            self._timeseries_id,
            instance_vals,
            instance_lookup_vector,
        )

    def set_sample_dimension(self, sample_dim: str):
        if self._sample_dimension != sample_dim:
            self._data = self._data.rename_dims({self._sample_dimension: sample_dim})
            self._sample_dimension = sample_dim
        return self._data

    def to_raster(self,
                  x_var,
                  y_var):
        return self._data.reset_index(self._timeseries_id)\
                         .set_index({self._instance_dimension: [x_var, y_var]})\
                         .unstack(self._instance_dimension)

    @staticmethod
    def _select_instances(
        ds: xr.Dataset,
        instance_dim: str,
        timeseries_id: str,
        instance_vals: Union[Sequence[Union[int, str]], np.ndarray, None] = None,
        instance_lookup_vector: Union[np.ndarray, None] = None,
    ) -> xr.Dataset:
        """
        Selects requested instances from an orthomulti timeseries array dataset.

        Returns a dataset containing the requested instances. If instances are requested
        that are not in the dataset, no error will be thrown.
        """
        if instance_lookup_vector is not None:
            instance_bool = instance_lookup_vector[ds[timeseries_id]]
        else:
            instance_bool = np.isin(ds[timeseries_id], instance_vals)
        return ds.sel({instance_dim: instance_bool})



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Test ASCAT Level 2 reader.
"""

import os
import sys
import pytest
import unittest
from datetime import datetime

import numpy as np
import numpy.testing as nptest

from ascat.read_native.bufr import AscatL2BufrFile
from ascat.read_native.nc import AscatL2NcFile
from ascat.eumetsat.level2 import AscatL2File
from ascat.eumetsat.level2 import AscatL2NcFileList
from ascat.eumetsat.level2 import AscatL2BufrFileList
from ascat.eumetsat.level2 import AscatL2EpsFileList

from get_path import get_testdata_path

eps_float_nan = -2147483648.
bufr_float_nan = 1.7e+38
uint8_nan = np.iinfo(np.uint8).max
uint16_nan = np.iinfo(np.uint16).max
float32_nan = -999999.


@pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")
class Test_AscatL2BufrFile(unittest.TestCase):

    def setUp(self):
        """
        Setup test files.
        """
        data_path = get_testdata_path(
        ) / "eumetsat/ASCAT_L2_SM_125/bufr/Metop_B"
        fname = data_path / 'M01-ASCA-ASCSMR02-NA-5.0-20170220050900.000000000Z-20170220055833-1207110.bfr'
        self.reader = AscatL2BufrFile(fname)

    def test_read(self):
        """
        Test read.
        """
        data, metadata = self.reader.read()

        ssm_should = np.array([
            29.2, 30.2, 35.7, 38.6, 37.5, 37.6, 40.5, 44.5, 40.7, 39.7, 41.5,
            38.8, 34.5, 36.8, 39.4, 41.2, 42.4, 42.9, 39.3, 30.5, 26.7, 26.5,
            26.7, 23.9, 26.2
        ])

        lats_should = np.array([
            64.74398, 64.81854, 64.89284, 64.96688, 65.04066, 65.11416,
            65.18739, 65.26036, 65.33304, 65.40545, 65.47758, 65.54942,
            65.62099, 65.69226, 65.76324, 65.83393, 65.90432, 65.97442,
            66.04422, 66.11371, 66.1829, 66.25177, 66.32034, 66.38859, 66.45653
        ])

        ssm_mean_should = np.array([
            36.7, 35.4, 33.4, 32.5, 32.5, 32., 31.2, 29.4, 28.7, 27.6, 25.8,
            25.4, 25.5, 25.3, 24.4, 23.4, 22.3, 21.3, 20.4, 20.4, 19.9, 19.7,
            20.3, 21.5, 22.9
        ])

        nptest.assert_allclose(data['lat'][:25], lats_should, atol=1e-5)
        nptest.assert_allclose(
            data['Surface Soil Moisture (Ms)'][:25], ssm_should, atol=1e-5)
        nptest.assert_allclose(
            data['Mean Surface Soil Moisture'][:25],
            ssm_mean_should,
            atol=1e-5)


class Test_AscatL2NcFile(unittest.TestCase):

    def setUp(self):
        """
        Setup test files.
        """
        data_path = get_testdata_path() / "eumetsat/ASCAT_L2_SM_125/nc/Metop_A"
        fname = data_path / 'W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,METOPA+ASCAT_C_EUMP_20170220041500_53652_eps_o_125_ssm_l2.nc'
        self.reader = AscatL2NcFile(fname)

    def test_read(self):
        """
        Test read.
        """
        data, metadata = self.reader.read()

        ssm_should = np.array([
            2.96000004, 0., 0., 0., 0., 0., 0., 0., 0., 1.82999992, 3.32999992,
            4.78999996, 4.31999969, 2.53999996, 0., 3.83999991, 5.76999998,
            1.5, 2.44000006, 4.11999989, 2.25999999, 2.65999985, 5.5999999,
            5.53999996, 4.85999966
        ])

        lats_should = np.array([
            62.60224, 62.67133, 62.74015, 62.80871, 62.877, 62.94502, 63.01276,
            63.08024, 63.14743, 63.21435, 63.28098, 63.34734, 63.41341,
            63.47919, 63.54468, 63.60988, 63.67479, 63.7394, 63.80372,
            63.86773, 63.93144, 63.99485, 64.05795, 64.12075, 64.18323
        ])

        ssm_mean_should = np.array([
            21.26000023, 21.27999878, 21.38999939, 22.43000031, 23.36999893,
            24.51000023, 26.01000023, 27.04999924, 26.94999886, 26.63999939,
            27.09999847, 27.56999969, 27.43000031, 26.64999962, 26.53999901,
            27.48999977, 28.20999908, 28.38999939, 28.79999924, 29.21999931,
            30.01000023, 30.97999954, 31.27999878, 31.8599987, 32.05999756
        ])

        nptest.assert_allclose(data['latitude'][:25], lats_should, atol=1e-5)
        nptest.assert_allclose(
            data['soil_moisture'][:25], ssm_should, atol=1e-5)
        nptest.assert_allclose(
            data['mean_soil_moisture'][:25], ssm_mean_should, atol=1e-5)

        data, metadata = self.reader.read(generic=True)
        nptest.assert_allclose(data['lat'][:25], lats_should, atol=1e-5)
        nptest.assert_allclose(
            data['sm'][:25], ssm_should, atol=1e-5)
        nptest.assert_allclose(
            data['sm_mean'][:25], ssm_mean_should, atol=1e-5)


@pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")
class Test_AscatL2NcFile_AscatL2BufrFile(unittest.TestCase):

    def setUp(self):
        """
        Setup test files.
        """
        data_path = get_testdata_path() / "eumetsat/ASCAT_L2_SM_125"

        fname_nc = data_path / "nc/Metop_A/W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,METOPA+ASCAT_C_EUMP_20170220041500_53652_eps_o_125_ssm_l2.nc"
        self.reader_nc = AscatL2NcFile(fname_nc)

        fname_bufr = data_path / "bufr/Metop_A/M02-ASCA-ASCSMR02-NA-5.0-20170220041500.000000000Z-20170220055656-1207110.bfr"
        self.reader_bufr = AscatL2BufrFile(fname_bufr)

    def test_read(self):
        """
        Test read.
        """
        data_nc, metadata = self.reader_nc.read()
        data_bufr, metadata = self.reader_bufr.read()

        nptest.assert_allclose(
            data_nc['latitude'], data_bufr['lat'], atol=1e-4)

        nc_bufr_matching = {
            'slope40':
                'Slope At 40 Deg Incidence Angle',
            'sigma40_error':
                'Estimated Error In Sigma0 At 40 Deg Incidence Angle',
            'utc_line_nodes':
                None,
            'wet_backscatter':
                'Wet Backscatter',
            'swath_indicator':
                None,
            'frozen_soil_probability':
                'Frozen Land Surface Fraction',
            'wetland_flag':
                'Inundation And Wetland Fraction',
            # The processing flag definition between BUFR and netCDF is slightly different
            # 'proc_flag1': 'Soil Moisture Processing Flag',
            'proc_flag2':
                None,
            'abs_line_number':
                None,
            'sat_track_azi':
                None,
            'sigma40':
                'Backscatter',
            'soil_moisture':
                'Surface Soil Moisture (Ms)',
            'soil_moisture_error':
                'Estimated Error In Surface Soil Moisture',
            # 'rainfall_flag': 'Rain Fall Detection',
            'soil_moisture_sensitivity':
                'Soil Moisture Sensitivity',
            'corr_flags':
                'Soil Moisture Correction Flag',
            'dry_backscatter':
                'Dry Backscatter',
            'aggregated_quality_flag':
                None,
            'mean_soil_moisture':
                'Mean Surface Soil Moisture',
            'as_des_pass':
                None,
            'slope40_error':
                'Estimated Error In Slope At 40 Deg Incidence Angle',
            'topography_flag':
                'Topographic Complexity',
            'snow_cover_probability':
                'Snow Cover'
        }

        # BUFR contains less accurate data so we only compare to 0.1
        for nc_name in nc_bufr_matching:

            bufr_name = nc_bufr_matching[nc_name]
            if bufr_name is None:
                continue

            if nc_name in ['mean_soil_moisture']:
                valid = ((data_nc[nc_name] != uint16_nan) &
                         (data_bufr[bufr_name] != bufr_float_nan))
            elif nc_name in [
                    'snow_cover_probability', 'rainfall_flag',
                    'topography_flag', 'frozen_soil_probability',
                    'wetland_flag', 'snow_cover_probability'
            ]:
                valid = ((data_nc[nc_name] != uint8_nan)
                         & (~np.isnan(data_nc[nc_name]))
                         & (data_bufr[bufr_name] != bufr_float_nan)
                         & (~np.isnan(data_bufr[bufr_name])))
            else:
                valid = ((data_nc[nc_name] != eps_float_nan)
                         & (~np.isnan(data_nc[nc_name]))
                         & (data_bufr[bufr_name] != bufr_float_nan)
                         & (~np.isnan(data_bufr[bufr_name])))

            nptest.assert_allclose(
                data_nc[nc_name][valid], data_bufr[bufr_name][valid], atol=0.1)


@pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")
class Test_AscatL2File(unittest.TestCase):

    def setUp(self):
        """
        Setup test files.
        """
        data_path = get_testdata_path() / "eumetsat/ASCAT_generic_reader_data"

        name_b = data_path / "bufr/M01-ASCA-ASCSMO02-NA-5.0-20180612035700.000000000Z-20180612044530-1281300.bfr"
        name_e = data_path / "eps_nat/ASCA_SMO_02_M01_20180612035700Z_20180612053856Z_N_O_20180612044530Z.nat"
        name_n = data_path / "nc/W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,METOPB+ASCAT_C_EUMP_20180612035700_29742_eps_o_250_ssm_l2.nc"

        self.bufr = AscatL2File(name_b)
        self.eps = AscatL2File(name_e)
        self.nc = AscatL2File(name_n)

    def test_read_all_formats(self):
        """
        Test read.
        """
        bufr_ds, metadata = self.bufr.read()
        eps_ds, metadata = self.eps.read()
        nc_ds, metadata = self.nc.read()

        for coord in ['lon', 'lat']:
            nptest.assert_allclose(bufr_ds[coord], eps_ds[coord], atol=1e-4)
            nptest.assert_allclose(eps_ds[coord], nc_ds[coord], atol=1e-4)
            nptest.assert_allclose(nc_ds[coord], bufr_ds[coord], atol=1e-4)

        matching = [
            'sm', 'sm_noise', 'sm_mean', 'sig40', 'sig40_noise', 'slope40',
            'slope40_noise', 'dry_sig40', 'wet_sig40', 'azi', 'sig', 'inc',
            'sm_sens', 'snow_prob', 'frozen_prob', 'wetland', 'topo', 'sat_id',
            'proc_flag', 'agg_flag', 'corr_flag', 'line_num', 'node_num',
            'sat_id', 'swath_indicator'
        ]

        # rounding issues in sat_track_azi leads to different as_des_pass
        # 'as_des_pass', 'sat_track_azi'

        # lists with no data fields
        nc_none = ['azi', 'inc', 'sig', 'corr_flag', 'proc_flag']

        # BUFR contain less accurate data so we only compare to 0.1
        for field in matching:

            # difference between the files should not be the case
            if field == 'sig40':
                mask = nc_ds[field] == float32_nan
                bufr_ds[field][mask] = float32_nan
                eps_ds[field][mask] = float32_nan

                nptest.assert_allclose(bufr_ds[field], eps_ds[field], atol=0.1)

            if field not in nc_none:
                nptest.assert_allclose(eps_ds[field], nc_ds[field], atol=0.1)
                nptest.assert_allclose(nc_ds[field], bufr_ds[field], atol=0.1)

    def test_eps(self):
        """
        Test read EPS.
        """
        eps_ds, metadata = self.eps.read()

        sm_should = np.array([
            69.11, 74.23, 74.12, 75.95, 76.23, 80.74, 83.45, 84.94, 84.28,
            86.33, 86.19, 86.31, 87.64, 87.92, 90.65, 90.52, 89.71, 89.33,
            91.41, 91.89, 94.51, 70.43, 67.75, 60.54, 69.43
        ])

        lat_should = np.array([
            64.06651, 64.21156, 64.355545, 64.49845, 64.64026, 64.78095,
            64.9205, 65.05891, 65.19613, 65.33216, 65.46697, 65.600555,
            65.73289, 65.86394, 65.9937, 66.12214, 66.249245, 66.374985,
            66.499344, 66.62231, 66.743835, 69.63313, 69.698105, 69.760895,
            69.821495
        ])

        lon_should = np.array([
            121.95572, 121.564156, 121.16849, 120.76867, 120.36467, 119.95644,
            119.54396, 119.12719, 118.70608, 118.2806, 117.85073, 117.41643,
            116.97765, 116.53439, 116.08661, 115.63427, 115.17735, 114.715836,
            114.24969, 113.77889, 113.30343, 96.66666, 96.049965, 95.42956,
            94.80551
        ])

        sm_mean_should = np.array([
            77.97, 77.57, 79.2, 78.38, 77.85, 79.81, 80.72, 81.23, 82.43,
            82.11, 81.93, 82.55, 83.41, 81.84, 81.43, 81.28, 80.37, 79.6,
            79.43, 78.02, 77.49, 42.42, 41.69, 42.99, 47.51
        ])

        t_should = np.array([
            '2018-06-12T03:56:59.999', '2018-06-12T03:56:59.999',
            '2018-06-12T03:56:59.999', '2018-06-12T03:56:59.999',
            '2018-06-12T03:56:59.999', '2018-06-12T03:56:59.999',
            '2018-06-12T03:56:59.999', '2018-06-12T03:57:03.750',
            '2018-06-12T03:57:03.750', '2018-06-12T03:57:03.750'
        ],
                            dtype='datetime64[ms]')

        nptest.assert_allclose(eps_ds['lat'][:25], lat_should, atol=1e-5)
        nptest.assert_allclose(eps_ds['lon'][:25], lon_should, atol=1e-5)
        nptest.assert_allclose(eps_ds['sm'][:25], sm_should, atol=1e-5)
        nptest.assert_allclose(
            eps_ds['sm_mean'][:25], sm_mean_should, atol=1e-5)
        nptest.assert_equal(eps_ds['time'][35:45], t_should)


@pytest.mark.skipif(sys.platform == 'win32', reason="Does not work on Windows")
class Test_AscatL2FileList(unittest.TestCase):
    """
    Test read AscatL2FileList in various formats.
    """

    def setUp(self):
        """
        Setup test data.
        """
        root_path = get_testdata_path() / "eumetsat/ASCAT_generic_reader_data"

        path = root_path / 'bufr'
        sat = 'b'
        product = 'smo'
        self.bufr_smo = AscatL2BufrFileList(path, sat, product)

        path = root_path / 'nc'
        sat = 'b'
        product = 'smo'
        self.nc_smo = AscatL2NcFileList(path, sat, product)

        path = root_path / 'eps_nat'
        sat = 'b'
        product = 'smo'
        self.eps_smo = AscatL2EpsFileList(path, sat, product)

    def test_smo_read_date(self):
        """
        Test read date for SMO formats.
        """
        dt = datetime(2018, 6, 12, 3, 57, 0)

        bufr_data, metadata = self.bufr_smo.read(dt)
        nc_data, metadata = self.nc_smo.read(dt)
        eps_data, metadata = self.eps_smo.read(dt)

        for coord in ['lon', 'lat']:
            nptest.assert_allclose(
                bufr_data[coord], eps_data[coord], atol=1e-4)
            nptest.assert_allclose(eps_data[coord], nc_data[coord], atol=1e-4)
            nptest.assert_allclose(nc_data[coord], bufr_data[coord], atol=1e-4)

    def test_smo_read_period(self):
        """
        Test read period for SMO formats.
        """
        dt_start = datetime(2018, 6, 12, 4, 0, 0)
        dt_end = datetime(2018, 6, 12, 4, 13, 0)

        date_field_fmt = "%Y%m%d%H%M%S"

        bufr_data, metadata = self.bufr_smo.read_period(
            dt_start, dt_end, date_field_fmt=date_field_fmt)

        nc_data, metadata = self.nc_smo.read_period(
            dt_start, dt_end, date_field_fmt=date_field_fmt)

        eps_data, metadata = self.eps_smo.read_period(
            dt_start, dt_end, date_field_fmt=date_field_fmt)

        for coord in ['lon', 'lat']:
            nptest.assert_allclose(
                bufr_data[coord], eps_data[coord], atol=1e-4)
            nptest.assert_allclose(eps_data[coord], nc_data[coord], atol=1e-4)
            nptest.assert_allclose(nc_data[coord], bufr_data[coord], atol=1e-4)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from pathlib import Path

import numpy as np
import xarray as xr

from pygeogrids.grids import genreg_grid
from pygeogrids.netcdf import load_grid, save_grid

from ascat.utils import dtype_to_nan


def retrieve_or_store_grid_lut(src_grid,
                               src_grid_id,
                               trg_grid_id,
                               trg_grid_size,
                               store_path=None):
    """
    Get a grid and its lookup table either from a store directory or
    create and return them.

    Parameters
    ----------
    src_grid : pygeogrids.BasicGrid
        Source grid.
    src_grid_id : str
        The source grid's id.
    trg_grid_id : str
        The target grid's id.
    trg_grid_size : int
        The size of the target grid in degrees.
    store_path : str, optional
        Path to the store directory (default: None).

    Returns
    -------
    trg_grid : pygeogrids.grids.BasicGrid
        Target grid.
    grid_lut : numpy.ndarray
        Look-up table.
    """
    if store_path == None:
        trg_grid = genreg_grid(trg_grid_size, trg_grid_size)
        grid_lut = trg_grid.calc_lut(src_grid).reshape(trg_grid.shape)
    else:
        store_path = Path(store_path)

        trg_grid_file = store_path / f"{trg_grid_id}.nc"

        if trg_grid_file.exists():
            trg_grid = load_grid(trg_grid_file)
        else:
            store_path.mkdir(parents=True, exist_ok=True)
            trg_grid = genreg_grid(trg_grid_size, trg_grid_size)
            save_grid(trg_grid_file, trg_grid)

        grid_lut_file = store_path / f"lut_{src_grid_id}_{trg_grid_id}.npy"

        if grid_lut_file.exists():
            grid_lut = np.load(grid_lut_file, allow_pickle=True)
        else:
            grid_lut = trg_grid.calc_lut(src_grid).reshape(trg_grid.shape)
            store_path.mkdir(parents=True, exist_ok=True)
            grid_lut.dump(grid_lut_file)

    return trg_grid, grid_lut


def regrid_swath_ds(ds, src_grid, trg_grid, grid_lut):
    """
    Convert a swath dataset to their nearest neighbors
    on a regular lat/lon grid.

    Parameters
    ----------
    ds : xarray.Dataset
        Swath dataset.
    src_grid : pygeogrids.grids.BasicGrid
        Sourde grid.
    trg_grid : pygeogrids.grids.BasicGrid
        Target grid.
    trg_grid_lut : numpy.ndarray
        Grid look-up table.

    Returns
    -------
    ds : xarray.Dataset
        Swath dataset resampled on a regular lat/lon grid.
    """
    index_lut = np.zeros(src_grid.n_gpi, dtype=np.int32) - 1
    index_lut[ds["location_id"].data] = np.arange(ds["location_id"].size)
    idx = index_lut[grid_lut]
    nan_pos = idx == -1

    coords = {
        "latitude": np.int32(trg_grid.lat2d[:, 0] / 1e-6),
        "longitude": np.int32(trg_grid.lon2d[0] / 1e-6)
    }

    regrid_ds = xr.Dataset(coords=coords)
    regrid_ds.attrs = ds.attrs

    regrid_ds["latitude"].attrs = ds["latitude"].attrs
    regrid_ds["longitude"].attrs = ds["longitude"].attrs
    dim = ("latitude", "longitude")

    for var in ds.variables:
        if var in ["latitude", "longitude"]:
            continue

        if ds[var].size == 1:
            continue

        regrid_ds[var] = (dim, ds[var].data[idx])
        regrid_ds[var].attrs = ds[var].attrs
        regrid_ds[var].encoding = {"zlib": True, "complevel": 4}

        if hasattr(ds[var], "_FillValue"):
            regrid_ds[var].data[nan_pos] = ds[var]._FillValue
        else:
            if var == "time":
                regrid_ds[var].data[nan_pos] = 0
            else:
                regrid_ds[var].data[nan_pos] = dtype_to_nan[ds[var].dtype]

    return regrid_ds



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import argparse
from pathlib import Path
from datetime import datetime

import re

from ascat.cell import CellGridFiles
from ascat.swath import SwathGridFiles

# based on https://stackoverflow.com/a/42865957/2002471
units = {"B": 1, "KB": 2**10, "MB": 2**20, "GB": 2**30, "TB": 2**40}


def parse_size(size):
    size = size.upper()
    #print("parsing size ", size)
    if not re.match(r' ', size):
        size = re.sub(r'([KMGT]?B)', r' \1', size)
    number, unit = [string.strip() for string in size.split()]
    return int(float(number) * units[unit])


class KeywordsAction(argparse.Action):

    def __call__(self, parser, namespace, values, option_string=None):
        keyword_dict = {}

        for arg in values:  #values => The args found for keyword_args
            pieces = arg.split('=')
            keyword_dict[pieces[0]] = pieces[1]

        setattr(namespace, self.dest, keyword_dict)


def parse_args_swath_stacker(args):
    """
    Parse command line arguments for stacking ASCAT swath files into a cell grid.

    Parameters
    ----------
    args : list
        Command line arguments.

    Returns
    -------
    parser : ArgumentParser
        Argument Parser object.
    """
    parser = argparse.ArgumentParser(
        description="Stack ASCAT swath files to a cell grid")
    parser.add_argument(
        "filepath",
        metavar="FILEPATH",
        type=str,
        help="Path to folder containing swath files")
    parser.add_argument(
        "outpath", metavar="OUTPATH", type=str, help="Path to the output data")
    parser.add_argument(
        "product_id",
        metavar="PRODUCT_ID",
        type=str,
        help="Product identifier")
    parser.add_argument(
        "--start_date",
        metavar="START_DATE",
        type=str,
        help="Start date in format YYYY-MM-DD. Must also provide end date if this is provided."
    )
    parser.add_argument(
        "--end_date",
        metavar="END_DATE",
        type=str,
        help="End date in format YYYY-MM-DD. Must also provide start date if this is provided."
    )
    parser.add_argument(
        "--dump_size",
        metavar="DUMP_SIZE",
        type=str,
        help="Size at which to dump the data to disk before reading more (default: 1GB)"
    )
    parser.add_argument(
        "--cells",
        metavar="CELLS",
        type=int,
        nargs='+',
        help="Numbers of the cells to process (default: None)")
    parser.add_argument(
        "--quiet",
        action="store_true",
        help="Do not print progress information")
    parser.add_argument(
        "fmt_kwargs",
        help="Format keyword arguments, depends on the product format used. Example: 'sat=A year=2008'",
        nargs='*',
        action=KeywordsAction)

    return parser.parse_args(args)


def swath_stacker_main(cli_args):
    """
    Main function for stacking ASCAT swath files into a cell grid given command line arguments.

    Parameters
    ----------
    cli_args : list
        Command line arguments.
    """
    args = parse_args_swath_stacker(cli_args)
    filepath = Path(args.filepath)
    product_id = args.product_id

    outpath = Path(args.outpath)
    outpath.parent.mkdir(parents=True, exist_ok=True)

    dump_size = parse_size(
        args.dump_size) if args.dump_size else parse_size("4GB")

    date_range = (args.start_date, args.end_date)
    if any(date_range) and not all(date_range):
        raise ValueError(
            "Both start and end date must be provided, or neither.")
    date_range = tuple((
        datetime.strptime(d, "%Y-%m-%d") for d in date_range if d is not None))
    date_range = None if not any(date_range) else date_range

    cells = args.cells

    quiet = args.quiet

    fmt_kwargs = args.fmt_kwargs

    swath_files = SwathGridFiles.from_product_id(filepath, product_id)
    if not quiet:
        print("Initializing...")
    swath_files.stack_to_cell_files(
        outpath,
        dump_size,
        date_range=date_range,
        fmt_kwargs=fmt_kwargs,
        cells=cells,
        print_progress=(not quiet),
    )


def parse_args_cell_format_converter(args):
    """
    Parse command line arguments for

    Parameters
    ----------
    args : list
        Command line arguments.

    Returns
    -------
    parser : ArgumentParser
        Argument Parser object.
    """
    parser = argparse.ArgumentParser(
        description="Reformat ASCAT cell files to various CF Array formats")
    parser.add_argument(
        "filepath",
        metavar="FILEPATH",
        type=str,
        help="Path to folder containing swath files")
    parser.add_argument(
        "outpath", metavar="OUTPATH", type=str, help="Path to the output data")
    parser.add_argument(
        "product_id",
        metavar="PRODUCT_ID",
        type=str,
        help="Product identifier")
    parser.add_argument(
        "arr_format",
        metavar="FORMAT",
        type=str,
        help="Output format (indexed, contiguous, or point.)")
    parser.add_argument(
        "--sf_format",
        metavar="SF_FORMAT",
        type=str,
    )

    return parser.parse_args(args)


def cell_format_converter_main(cli_args):

    args = parse_args_cell_format_converter(cli_args)
    filepath = Path(args.filepath)
    product_id = args.product_id
    outpath = Path(args.outpath)
    outpath.parent.mkdir(parents=True, exist_ok=True)
    array_format = args.arr_format
    sf_format = args.sf_format

    cell_files = CellGridFiles.from_product_id(
        filepath, product_id, sf_format=sf_format)

    if array_format == "contiguous":
        cell_files.convert_to_contiguous(outpath)
    elif array_format == "indexed":
        cell_files.reprocess(outpath, lambda ds: ds, ra_type="indexed")
    elif array_format == "point":
        cell_files.reprocess(outpath, lambda ds: ds, ra_type="point")
    else:
        raise ValueError(f"Invalid array format: {array_format}")


def run_swath_stacker():
    """Run command line interface for temporal aggregation of ASCAT data."""
    swath_stacker_main(sys.argv[1:])


def run_cell_format_converter():
    """Run command line interface for temporal aggregation of ASCAT data."""
    cell_format_converter_main(sys.argv[1:])
